#!/usr/bin/python
#coding:utf-8

"""
@author: yangst2
@contact: yangst2@lenovo.com
@software: PyCharm
@file: utils.py
@time: 2020/4/11 19:48
"""

import re
import numpy as np
from typing import Tuple, List


def sequence_padding(inputs, length=None, padding=0):
    if length is None:
        length = max([len(x) for x in inputs])
    pad_width = [(0, 0) for _ in np.shape(inputs[0])]
    outputs = []
    for x in inputs:
        x = x[:length]
        pad_width[0] = (0, length - len(x))
        x = np.pad(x, pad_width, 'constant', constant_values=padding)
        outputs.append(x)
    return np.array(outputs)


def remove_multi_blank(text):
    return re.sub(r"\s+", " ", text).lstrip().rstrip()


def read_plain_text(file_path, split_tag=None) -> Tuple[List[str], List[str]]:
    data_x, data_y = [], []
    with open(file_path, encoding="utf8") as f:
        for line in f.readlines()[0:]:
            line = line.strip("\n")
            if split_tag is None:
                split_tag = "\t"
            parts = line.split(split_tag)
            if len(parts) == 2:
                data_x.append(parts[0])
                data_y.append(str(parts[1]).lower())
    return list(data_x), list(data_y)


def build_mapping(label_vocab, add_one=False):
    label2id = dict()
    id2label = dict()
    for index, label in enumerate(label_vocab):
        if add_one:
            index += 1
        label2id[label] = index
        id2label[index] = label
    return label2id, id2label

def main():

    train_file_path = "../demo_data/train.txt"
    test_file_path = "../demo_data/train.txt"

    import os

    print(os.path.split(train_file_path))

    x, y = read_plain_text(train_file_path)
    for i in range(len(x)):
        print(x[i], y[i])


if __name__ == "__main__":
    main()


#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Created by iFantastic on 2020/5/6

from typing import List


def build_label_vocab(labels: List[str], multi_label=False):
    """
    根据标签值，构造标签集合
    :param labels: 标签序列
    :param multi_label: 是否为多标签
    :return:
    """
    label_vocab = set()
    if not multi_label:
        label_vocab = set(labels)
    else:
        for ll in labels:
            ll = ll.split("|")
            label_vocab.update(ll)
    return label_vocab


def to_category(row_labels: str, label2id: dict, multi_label=False):
    tmp = [0] * len(label2id)
    if not multi_label:
        tmp[label2id[row_labels]] = 1
        return tmp
    else:
        for l in row_labels.split("|"):
            tmp[label2id[l]] = 1
        return tmp




#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Created by iFantastic on 2020/4/16

from tornado.escape import json_decode
from tornado.web import RequestHandler
from lenluhub.tasks.classification_task import Classifier, BertClassifier
from lenluhub.processor.tokenizer import BertTokenizer
from lenluhub.common.errors import ErrorMessage
from typing import List
from abc import ABC

# 为了解决一个bot需要多个服务的情况，这里将服务和handlder进行绑定，即一个handler 这样新增model服务的时候 只需要增加一个hanlder
# 然后所有的handlder注册到一个server下面


class SingleInputHandler(RequestHandler, ABC):

    def __init__(self, application, request):
        super().__init__(application, request)

    def set_predict_model(self, model: Classifier, tokenizer: BertTokenizer):
        self.model = model
        self.tokenizer = tokenizer

    def predict_label(self, sentence):
        """
        :param sentence:
        :return:
        """
        enc = self.tokenizer.encode(sentence)
        result = self.model.predict([enc[0], enc[1]])
        return result

    def post(self, *args, **kwargs):
        """
        :param args:
        :param kwargs:
        :return:
        """
        try:
            data_rec = json_decode(self.request.body)
            if data_rec is None or len(data_rec) == 0 or "query" not in data_rec.keys():
                self.write(ErrorMessage.reques_parameter_error)
            else:
                source_sentence = data_rec["query"]
                result = self.predict_label(source_sentence)
                self.write(result)
        except Exception as e:
            print(e)
            self.write(ErrorMessage.server_inner_error)


if __name__ == '__main__':

    pass


#!/usr/bin/python
#coding:utf-8

"""
@author: yangst2
@contact: yangst2@lenovo.com
@software: PyCharm
@file: metrics.py
@time: 2020/4/12 20:39
"""

from sklearn.metrics import f1_score, accuracy_score, recall_score
import tensorflow.keras as keras
import numpy as np
import json
import os


class MultiClassCallback(keras.callbacks.Callback):
    best_f1 = 0.0
    best_epoch = 0.0

    def __init__(self, model_save_dir, patience=5, test_data=None, id2label=None):
        self.model_save_dir = model_save_dir
        self.patience = patience
        self.test_data = test_data
        self.id2label = id2label
        self.metrics_score = dict()
        self.metrics_score.setdefault("f1", [])
        self.metrics_score.setdefault("precision", [])
        self.metrics_score.setdefault("recall", [])
        super().__init__()

    def on_epoch_end(self, epoch, logs=None):
        predict_result = self.model.predict(self.test_data[0])
        pred_label = [self.id2label[_] for _ in np.argmax(predict_result, -1)]
        real_label = [self.id2label[_] for _ in np.argmax(self.test_data[1], -1)]
        f1 = f1_score(real_label, pred_label, average="micro")
        precision = accuracy_score(real_label, pred_label)
        recall = recall_score(real_label, pred_label, average="micro")
        self.metrics_score["f1"].append(f1)
        self.metrics_score["precision"].append(precision)
        self.metrics_score["recall"].append(recall)
        print(" - Current Epoch: {} Validate F1: {} and the Best Validate F1:{}".format(epoch, f1, self.best_f1))
        if f1 > self.best_f1:
            self.best_f1 = f1
            self.best_epoch = epoch
            self.model.save_weights(self.model_save_dir + '/best_model.weights')
        else:
            if epoch - self.best_epoch >= self.patience:
                self.model.stop_training = True
                print("Early stopping at epoch {}, and final Validate F1: {}".format(epoch, self.best_f1))

    def on_train_end(self, logs=None):
        with open(os.path.join(self.model_save_dir, "train_log.txt"), "w", encoding="utf8") as f:
            f.write(json.dumps(self.metrics_score))



#!/usr/bin/env python
# encoding: utf-8

"""
@version: v1.0
@author: YangST
@software: PyCharm Community Edition
@file: cnn_online.py
@time: 2017/8/30 9:43
"""
from abc import ABC

import logging
import asyncio
import tornado.web
import tornado.ioloop
import tornado.options
from tornado.options import define, options
from tornado.web import RequestHandler
from typing import Type
from lenluhub.web.handlers import SingleInputHandler


class Service:

    def __init__(self):
        pass



class HttpsService(Service):

    def __init__(self, handler: RequestHandler, service_port=9001, service_url=""):
        tornado.options.parse_command_line()
        self.service_port = service_port
        self.service_url = service_url
        self.handler = handler
        define("port", default=self.service_port, help="run on the given port", type=int)
        self.app = tornado.web.Application([(r"/{}".format(self.service_url), self.handler)])
        self.app.listen(options.port)
        self.__ioloop = tornado.ioloop.IOLoop.instance()
        super().__init__()

    def start_service(self):
        self.__ioloop.start()

    def stop_service(self):
        self.__ioloop.stop()




#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Created by iFantastic on 2020/4/16


class ErrorMessage:

    reques_parameter_error = "Request Parameter Error..."
    server_inner_error = "Server Inner Error..."




#!/usr/bin/python
#coding:utf-8

"""
@author: yangst2
@contact: yangst2@lenovo.com
@software: PyCharm
@file: Tokenizer.py
@time: 2020/4/11 19:50
"""

import json
import shutil, os
from collections import defaultdict
from lenluhub.common.utils import remove_multi_blank, sequence_padding
from bert4keras.tokenizers import Tokenizer


class BertTokenizer(Tokenizer):

    def __init__(self, config: defaultdict):
        self.config = config
        self.token_dict_path = config.get("token_dict_path", None)
        self.max_len = config.get("max_len", 512)
        self.lower_case = config.get("lower_case", False)
        self.remove_multi_blank = config.get("remove_multi_blank", False)
        super().__init__(self.token_dict_path)

    def pre_process(self, text):
        """针对训练样本进行预处理"""
        if text is None:
            return None
        if self.lower_case:
            text = text.lower()
        if self.remove_multi_blank:
            text = remove_multi_blank(text)
        return text

    def encode(
        self,
        first_text,
        second_text=None,
        max_length=None,
        first_length=None,
        second_length=None
    ):
        first_text = self.pre_process(first_text)
        if second_text is not None:
            second_text = self.pre_process(second_text)
        enc = super().encode(first_text=first_text, second_text=second_text, max_length=self.max_len)
        enc = sequence_padding(enc, length=self.max_len)
        return enc[0].tolist(), enc[1].tolist()

    def save_config_to_file(self, save_path):
        """
        保存配置，便于重新加载部署服务
        :param save_path:
        :return:
        """
        with open("{}/tokenizer_config.json".format(save_path), "w", encoding="utf8") as f:
            print(self.token_dict_path)
            (_, config_file_name) = os.path.split(self.token_dict_path)
            shutil.copyfile(self.token_dict_path, dst=os.path.join(save_path, config_file_name))
            self.config["token_dict_path"] = config_file_name
            content = json.dumps(self.config, ensure_ascii=False)
            f.write(content+"\n")

    @staticmethod
    def load(save_path: str):
        """
        根据Config加载得到Tokenizer
        :param save_path:
        :return:
        """
        with open("{}/tokenizer_config.json".format(save_path), "r", encoding="utf8") as f:
            config = json.loads(f.read())
            if "token_dict_path" in config.keys():
                config["token_dict_path"] = os.path.join(save_path, config["token_dict_path"])
            return BertTokenizer(config)


if __name__ == '__main__':

    config = defaultdict()
    config["token_dict_path"] = "../../en_bert_resource_uncased/vocab.txt"
    config["max_len"] = 10
    config["lower_case"] = False
    config["remove_multi_blank"] = False

    tokenizer = BertTokenizer(config)
    print(tokenizer.encode("i am a boyd"))




#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Created by iFantastic on 2020/4/17

import logging
import asyncio
import tornado.web
import tornado.ioloop
import tornado.options
from tornado.options import define, options
from tornado.web import RequestHandler
from typing import Type
from lenluhub.web.handlers import SingleInputHandler
from tornado.escape import json_decode
from lenluhub.common.errors import ErrorMessage
from lenluhub.processor.tokenizer import BertTokenizer
from lenluhub.tasks.classification_task import BertClassifier
from lenluhub.web.service import HttpsService
import asyncio

class MainHandler(RequestHandler):

    def __init__(self, application, request):
        super().__init__(application, request)

    def post(self, *args, **kwargs):
        param_format = """
        {
           "query":"how to update my phone"
        }
        """
        print(self.request.body)

        data_rec = json_decode(self.request.body)
        print(data_rec)
        if data_rec is None or len(data_rec) == 0 or "bot_id" not in data_rec.keys() or "task_name" not in data_rec.keys():
            self.write(ErrorMessage.reques_parameter_error)
        else:
            task_name = data_rec["task_name"]
            service_url = ""
            if "service_url" in data_rec.keys():
                service_url = data_rec["service_url"]
            model_save_dir = "../../demo/demo_model"
            tokenizer = BertTokenizer.load(model_save_dir)
            model = BertClassifier.load_model(model_save_dir)
            hanlder = type("{}_service_handler".format(task_name), (SingleInputHandler,), {"model": model, "tokenizer": tokenizer})
            self.application.add_handlers(r".*", [("/{}".format(service_url), hanlder)])
            self.write("模型训练完成，服务已启动，请访问:http://127.0.0.1:{}/{}\n进行验证，访问参数要求：\n{}".format(9001, service_url, param_format))

if __name__ == "__main__":

    app = tornado.web.Application(handlers=[(r"/", MainHandler)])
    http_server = tornado.httpserver.HTTPServer(app)
    http_server.listen(9001)
    tornado.ioloop.IOLoop.instance().start()



#!/usr/bin/python
#coding:utf-8

"""
@author: yangst2
@contact: yangst2@lenovo.com
@software: PyCharm
@file: ClassificationDataset.py
@time: 2020/4/12 9:45
"""

import logging
import numpy as np
from lenluhub.common.utils import read_plain_text, build_mapping
from lenluhub.processor.tokenizer import BertTokenizer
from lenluhub.dataset import dataset_utils


class BertClsDataSet:

    def __init__(self, train_data_path, test_data_path=None,
                 tokenizer: BertTokenizer = None, multi_label=False,
                 max_training_case_num=10):
        """
        :param train_data_path: 训练数据路径
        :param test_data_path: 测试数据路径
        :param tokenizer: 分词器
        :param multi_label: 是否为多标签
        """
        self.train_data_path = train_data_path
        self.test_data_path = test_data_path
        self.tokenizer = tokenizer
        self.multi_label = multi_label
        self.max_training_case_num = max_training_case_num

        # 以下属性仅初始化，通过加载数据重新赋值
        self.train_data, self.test_data = None, None
        self.train_case_num, self.test_case_num, self.label_num = 0.0, 0.0, 0.0
        self.label2id, self.id2label = None, None
        self.load_data_set(train_data_path, test_data_path)
        self.label_vocab = None
        self.feature_num = 0

    def load_data_set(self, train_data_path, test_data_path):
        """
        处理数据集
        """
        logging.info("self: " + str(self))
        train_data_x, train_data_y = read_plain_text(train_data_path, split_tag="\t")
        test_data_x, test_data_y = read_plain_text(test_data_path, split_tag="\t")
        train_token_x, train_segment_x, test_token_x, test_segment_x = [], [], [], []
        for row in train_data_x:
            row_enc = self.tokenizer.encode(row)
            train_token_x.append(row_enc[0])
            train_segment_x.append(row_enc[1])
        train_token_x = np.array(train_token_x)
        train_segment_x = np.array(train_segment_x)
        for row in test_data_x:
            row_enc = self.tokenizer.encode(row)
            test_token_x.append(row_enc[0])
            test_segment_x.append(row_enc[1])
        test_token_x = np.array(test_token_x)
        test_segment_x = np.array(test_segment_x)

        train_label_vocab = dataset_utils.build_label_vocab(train_data_y, multi_label=self.multi_label)

        tmp_train_data_y, tmp_test_data_y = [], []
        label2id, id2label = build_mapping(train_label_vocab)
        for row_labels in train_data_y:
            tmp_train_data_y.append(dataset_utils.to_category(row_labels, label2id=label2id, multi_label=self.multi_label))
        for row_labels in test_data_y:
            tmp_test_data_y.append(dataset_utils.to_category(row_labels, label2id=label2id, multi_label=self.multi_label))

        train_token_x = np.array(train_token_x)[:self.max_training_case_num]
        train_segment_x = np.array(train_segment_x)[:self.max_training_case_num]
        train_data_y = np.array(tmp_train_data_y)[:self.max_training_case_num]
        test_token_x = np.array(test_token_x)[:self.max_training_case_num]
        test_segment_x = np.array(test_segment_x)[:self.max_training_case_num]
        test_data_y = np.array(tmp_test_data_y)[:self.max_training_case_num]

        self.train_data = [[train_token_x, train_segment_x], train_data_y]
        self.test_data = [[test_token_x, test_segment_x], test_data_y]
        self.label_vocab = train_label_vocab
        self.label_num = len(train_label_vocab)
        self.id2label = id2label




#!/usr/bin/python
#coding:utf-8

"""
@author: yangst2
@contact: yangst2@lenovo.com
@software: PyCharm
@file: ClassificationTask.py
@time: 2020/4/12 11:59
"""

import json, os, shutil
import tensorflow.keras as keras
from lenluhub.dataset.classification_dataset import BertClsDataSet
from bert4keras.models import build_transformer_model
from lenluhub.evaluate import metrics
from collections import defaultdict
import keras


class Classifier:

    def __init__(self):
        pass

    def predict(self, tmp_input, top_k=1):
        raise NotImplementedError


class BertClassifier(Classifier):

    def __init__(self, config: defaultdict):
        self.task_name = config["task_name"]
        self.bot_id = config["bot_id"]
        self.model_save_dir = config["model_save_dir"]
        self.id2label = config.get("id2label", None)
        self.multi_label = config.get("multi_label", False)
        self.max_len = config.get("max_len", 0)
        self.label_num = config.get("label_num", 0)
        self.bert_config_path = config.get("bert_config_path", None)
        self.bert_checkpoint_path = config.get("bert_checkpoint_path", None)
        self.train_data, self.test_data = None, None
        self.tokenizer = None
        self.model = None
        super().__init__()

    def load_config_from_data_set(self, data_set: BertClsDataSet):
        """根据DataSet确定构建模型所需要的部分参数"""
        self.id2label = data_set.id2label
        self.max_len = data_set.tokenizer.max_len
        self.multi_label = data_set.multi_label
        self.label_num = data_set.label_num
        self.train_data = data_set.train_data
        self.test_data = data_set.test_data

    def build_model(self):
        bert = build_transformer_model(
            config_path=self.bert_config_path,
            checkpoint_path=self.bert_checkpoint_path,
            model='bert',
            return_keras_model=False
        )
        output = keras.layers.Lambda(lambda x: x[:, 0])(bert.model.output)
        output = keras.layers.Dense(units=self.label_num, activation=keras.activations.softmax,)(output)
        self.model = keras.models.Model(bert.model.inputs, output)
        self.model.compile(loss=keras.losses.categorical_crossentropy, optimizer=keras.optimizers.Adam(lr=0.0001), metrics=["acc"])

    def fit(self, batch_size=64, epochs=15):
        callback = metrics.MultiClassCallback(self.model_save_dir, test_data=self.test_data, id2label=self.id2label)
        train_history = self.model.fit(x=self.train_data[0], y=self.train_data[1], batch_size=batch_size, epochs=epochs, callbacks=[callback])
        return train_history

    def predict(self, tmp_input, top_k=1):
        probability = self.model.predict([[tmp_input[0]], [tmp_input[1]]])[0]
        label_props = dict()
        for i in range(len(probability)):
            label_props[self.id2label[i]] = probability[i]
        items = [item for item in label_props.items()]
        items.sort(key=lambda x: x[1], reverse=True)
        result = []
        for item in items[:top_k]:
            tmp = dict()
            tmp[self.task_name] = item[0]
            tmp["confidence"] = float(item[1])
            tmp["procMethod"] = "model"
            result.append(tmp)
        return json.dumps(result, ensure_ascii=False)

    def save(self, model_save_dir):
        params = dict()
        params["bot_id"] = self.bot_id
        params["task_name"] = self.task_name
        params["max_len"] = self.max_len
        params["label_num"] = self.label_num
        params["multi_label"] = self.multi_label
        params["id2label"] = self.id2label
        (_, config_file_name) = os.path.split(self.bert_config_path)
        shutil.copyfile(self.bert_config_path, dst=os.path.join(model_save_dir, config_file_name))
        params["bert_config_path"] = config_file_name
        params["model_save_dir"] = self.model_save_dir
        with open("{}/model_config.json".format(model_save_dir), "w", encoding="utf8") as f:
            f.write(json.dumps(params, ensure_ascii=False)+"\n")

    @staticmethod
    def load_model(save_path):
        """
        根据Config自动创建模型，加载相应的权重
        :param model_config_path:
        :return:
        """
        with open("{}/model_config.json".format(save_path), "r", encoding="utf8") as f:
            params = json.loads(f.read())
            new_id2label = dict()
            if "id2label" in params.keys():
                for item in params["id2label"].items():
                    new_id2label[int(item[0])] = item[1]
            params["id2label"] = new_id2label
            if "bert_config_path" in params.keys():
                params["bert_config_path"] = os.path.join(save_path, params["bert_config_path"])
        model = BertClassifier(params)
        model.build_model()
        model.model.load_weights("{}/best_model.weights".format(save_path))
        return model

    def summary(self):
        self.model.summary()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Created by iFantastic on 2020/5/6


from setuptools import setup, find_packages
from distutils.core import setup
import sys

if sys.version_info < (3, 6):
    sys.exit('Python 3.6 or greater is required.')

with open('README.md', 'r') as fp:
    readme = fp.read()

# 版本号，自己随便写
VERSION = "0.0.3"

LICENSE = "MIT"

setup(
    name='lenluhub',
    version=VERSION,
    description=(
        'lenluhub'
    ),
    long_description=readme,
    author='yangst2',
    author_email='yangst2@lenovo.com',
    maintainer='yanst2',
    maintainer_email='yangst2@lenovo.com',
    license=LICENSE,
    packages=find_packages(),
    url="https://gitlab.lenovo.com/moli",
    platforms=["all"],
    install_requires=[
        "tensorflow==2.1.0",
        "keras==2.3.1",
        "bert4keras==0.7.5",
        "tornado==6.0.4"
    ],
    classifiers=[
        'Development Status :: 4 - Beta',
        'Operating System :: OS Independent',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python',
        'Programming Language :: Python :: Implementation',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Topic :: Software Development :: Libraries'
    ],
)
